﻿using System;
using Ozeki.VoIP;
using Ozeki.Media.MediaHandlers;
using Ozeki.Media;

namespace CallTransfer
{
    public class Transfer
    {
        private TransferState state;
        private readonly IPhoneCall call;
        private readonly IPhoneCall transferCall;
        private readonly MediaConnector connector;
        private readonly PhoneCallAudioSender callSender;
        private readonly PhoneCallAudioReceiver callReceiver;
        private readonly PhoneCallAudioSender transferCallSender;
        private readonly PhoneCallAudioReceiver transferCallReceiver;

        public event EventHandler<VoIPEventArgs<TransferState>> TransferStateChanged; 

        public Transfer(IPhoneCall call, IPhoneCall transferCall)
        {
            this.call = call;
            this.transferCall = transferCall;

            state = TransferState.Idle;

            // subscribe for call events
            transferCall.CallStateChanged += (Call_CallStateChanged);
            transferCall.CallErrorOccured += (Call_CallErrorOccured);
            call.CallStateChanged += (Call_CallStateChanged);
            call.CallErrorOccured += (Call_CallErrorOccured);

            // create media handlers
            callSender = new PhoneCallAudioSender();
            callReceiver = new PhoneCallAudioReceiver();
            transferCallSender = new PhoneCallAudioSender();
            transferCallReceiver = new PhoneCallAudioReceiver();

            // connect media handlers
            connector = new MediaConnector();
            connector.Connect(callReceiver, transferCallSender);
            connector.Connect(transferCallReceiver, callSender);

            // attach media handlers to calls
            callSender.AttachToCall(call);
            callReceiver.AttachToCall(call);
            transferCallSender.AttachToCall(transferCall);
            transferCallReceiver.AttachToCall(transferCall);
        }

        /// <summary>
        /// Begins the transfer.
        /// </summary>
        public void Begin()
        {
            call.Hold();
            transferCall.Start();
            SetState(TransferState.Transferring);
        }

        /// <summary>
        /// Ends the transfer.
        /// </summary>
        public void End()
        {
            StopCall();
            SetState(TransferState.Ended);
        }

        private void StopCall()
        {
            // hang up calls, unsubscribe from call events
            if (call != null)
            {
                call.HangUp();
                call.CallStateChanged -= (Call_CallStateChanged);
                call.CallErrorOccured -= (Call_CallErrorOccured);  
            }

            if (transferCall != null)
            {
                transferCall.HangUp();
                transferCall.CallStateChanged -= (Call_CallStateChanged);
                transferCall.CallErrorOccured -= (Call_CallErrorOccured);
            }

            // detach media handlers from calls
            callSender.Detach();
            callReceiver.Detach();
            transferCallSender.Detach();
            transferCallReceiver.Detach();

            // dispose media handlers
            connector.Dispose();
            callSender.Dispose();
            callReceiver.Dispose();
            transferCallSender.Dispose();
            transferCallReceiver.Dispose();
        }

        private void Call_CallStateChanged(object sender, VoIPEventArgs<CallState> e)
        {
            // rejected or busy
            if (e.Item == CallState.Busy || e.Item == CallState.Rejected)
            {
                SetState(TransferState.Rejected);
                return;
            }

            // transfer error
            if (e.Item == CallState.Error)
            {
                SetState(TransferState.Error);
                return;
            }

            // transfer call picked up
            if (e.Item.IsInCall())
            {
                if (call == null)
                {
                    transferCall.HangUp();
                    return;
                }

                // transfer successful
                if (state == TransferState.Transferring)
                {
                    SetState(TransferState.Transferred);
                    call.Hold();
                }

                return;
            }

            // transfer ended
            if (e.Item.IsCallEnded())
            {
                StopCall();

                if (state == TransferState.Idle || state == TransferState.Transferring)
                    SetState(TransferState.Error);
                else
                    SetState(TransferState.Ended);

                return;
            }
        }

        private void Call_CallErrorOccured(object sender, VoIPEventArgs<CallError> e)
        {
            StopCall();
            SetState(TransferState.Error);
        }

        private void SetState(TransferState newState)
        {
            state = newState;
            if (TransferStateChanged != null)
                TransferStateChanged(this, new VoIPEventArgs<TransferState>(state));
        }
    }
}
